import 'package:flutter/widgets.dart';
import 'package:microfrontends/src/resolver.dart';

import 'async.dart';
import 'effect.dart';
import 'state_container_provider.dart';
import 'mutation.dart';
import 'query.dart';
import 'state_container.dart';
import 'widgets.dart';

@protected
extension Queries on BuildContext {
  @protected
  StateContainerProviderElement? getContainerElement() =>
      this.getElementForInheritedWidgetOfExactType<StateContainerProvider>()
          as StateContainerProviderElement?;

  StateContainer<T> getContainerOfType<T>() {
    try {
      final el = getContainerElement();
      final container = el?.containers[T];
      return container! as StateContainer<T>;
    } catch (err) {
      throw new Exception('No StateContainer<$T> found');
    }
  }

  Resolver getResolverForType<R>(Type type) {
    final el = getContainerElement();
    Resolver? r;

    if (R == QueryResolver) {
      r = el?.queryResolvers[type];
    } else {
      r = el?.mutationResolvers[type];
    }

    if (r == null) {
      throw new Exception('No resolver found for $type');
    }

    return r;
  }

  QueryResolver getResolverForQueryType<T>() {
    final r = getResolverForType<QueryResolver>(T);
    return r as QueryResolver;
  }

  MutationResolver getMutationResolverOfType(Type type) {
    final r = getResolverForType<MutationResolver>(type);
    return r as MutationResolver;
  }

  StateContainerProviderElement getContainerHolder(container) {
    final el = getContainerElement();
    return el!.elements[container.runtimeType]!;
  }
}

extension FrameworkOps on BuildContext {
  T? query<T>([Object? params]) {
    return Query<T>(params).exec(this);
  }

  Widget subscribe<T>({
    Object? params,
    Widget? child,
    required SubscriptionBuilder<T> builder,
  }) {
    return QueryBuilder<T>(
      query: Query<T>(params),
      builder: builder,
      child: child,
    );
  }

  Widget subscribeAsync<T>({
    Object? params,
    Widget? child,
    required SubscriptionBuilder<Async<T>> builder,
  }) {
    return subscribe<Async<T>>(params: params, child: child, builder: builder);
  }

  void mutation<T extends Mutation>(T mutation) {
    mutation.exec(this);
  }

  Effect<T> effect<T>(EffectCallback<T> callback) {
    return Effect<T>(callback);
  }
}

extension FrameworkWidgets on Widget {
  Widget withContainer<T>(
    BuildContext context, [
    StateContainer<T>? container,
  ]) {
    StateContainer<T> c = container ?? StateContainer.of<T>(context);

    return StateContainerProvider(
      container: c,
      child: this,
    );
  }
}
