import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:microfrontends/microfrontends.dart';

import 'resolver.dart';
import 'extensions.dart';

typedef Predicate<T> = bool Function(T a, T b);
final Predicate _neq = (a, b) => a != b;

abstract class UpdatePolicy {}

class Rebuild extends UpdatePolicy {
  final bool shouldRebuild;

  Rebuild(this.shouldRebuild);
}

class NotifyListeners<T> extends UpdatePolicy {
  final bool shouldNotify;
  final T payload;

  NotifyListeners(this.shouldNotify, this.payload);

  notify(BuildContext context) {
    Effect.dispatch<T>(context, payload);
  }
}

class QueryAspect<T> {
  final Query<T> query;
  final Resolver resolver;
  T? _cachedResult;

  Type get type => T;

  QueryAspect(this.query, this.resolver);

  UpdatePolicy resolveUpdatePolicy(ValueNotifier stateNotifier) {
    final newValue = resolver.resolve(stateNotifier, query.params);
    final didChange = query.updateCriteria(_cachedResult, newValue);
    _cachedResult = newValue;

    return Rebuild(didChange);
  }
}

class OffstageQueryAspect<T> extends QueryAspect<T> {
  OffstageQueryAspect(Query<T> query, Resolver resolver)
      : super(query, resolver);

  @override
  UpdatePolicy resolveUpdatePolicy(ValueNotifier stateNotifier) {
    final p = super.resolveUpdatePolicy(stateNotifier) as Rebuild;
    return NotifyListeners<T>(p.shouldRebuild, _cachedResult!);
  }
}

class Query<T> {
  final Object? params;
  final Predicate updateCriteria;

  Query([this.params, Predicate? updateCriteria])
      : this.updateCriteria = updateCriteria ?? _neq;

  T? exec(BuildContext context) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    return r.resolve(container.stateNotifier, params);
  }

  T? subscribe(BuildContext context, {bool offstage = false}) {
    final r = context.getResolverForQueryType<T>();
    final container = r.getContainer(context);
    final element = context.getContainerHolder(container);

    final result = r.resolve(container.stateNotifier, params);
    final aspect =
        offstage ? OffstageQueryAspect(this, r) : QueryAspect(this, r);

    aspect._cachedResult = result;

    context.dependOnInheritedElement(element, aspect: aspect);

    return result;
  }

  unsubscribe(Element element) {
    final r = element.getResolverForQueryType<T>();
    final container = r.getContainer(element);
    final el = element.getContainerHolder(container);
    el.unsubscribe(element);
  }

  Query<T> withUpdateCondition(bool Function(T state) condition) {
    return Query<T>(params, (oldState, newState) {
      final shouldNotify = updateCriteria(oldState, newState);
      final matchesCondition = condition(newState);
      return shouldNotify && matchesCondition;
    });
  }
}

class QueryResolver<T, K> extends Resolver<T, K, Object> {
  Type get keyType => K;

  QueryResolver(
      ResolverController<T> ctrl, ResolverCallback<K, T, Object> callback)
      : super(ctrl, callback);
}
