import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:microfrontends/src/effect.dart';

import 'context.dart';
import 'mutation.dart';
import 'query.dart';
import 'state_container.dart';
import 'extensions.dart';
import 'state_container_context.dart';

class _Update {
  final List<Element> dirty = [];
  final List<Element> inactive = [];
  final List<MapEntry<Element, NotifyListeners>> daemons = [];
}

class StateContainerProvider extends InheritedWidget {
  final StateContainer container;

  StateContainerProvider({
    Key? key,
    required this.container,
    required Widget child,
  }) : super(key: key, child: child);

  @override
  bool updateShouldNotify(StateContainerProvider oldWidget) {
    return container.state != oldWidget.container.state;
  }

  @override
  InheritedElement createElement() {
    return StateContainerProviderElement(this);
  }
}

class StateContainerProviderElement extends InheritedElement
    implements StateContainerContext {
  final StateContainer container;

  StateContainerProviderElement(StateContainerProvider widget)
      : container = widget.container,
        super(widget);

  final containers = Context<Type, StateContainer>();
  final queryResolvers = Context<Type, QueryResolver>();
  final mutationResolvers = Context<Type, MutationResolver>();
  final elements = Context<Type, StateContainerProviderElement>();
  final dependencyAspects = Context<Element, QueryAspect>();

  Set<Element> _subscribers = {};
  StreamSubscription? _initSub;

  K? query<K>([Object? args]) {
    return Query<K>(args).exec(this);
  }

  @override
  void mount(Element? parent, newSlot) {
    _initSub = container.init(this).listen((event) {
      container.state = event;
    });

    container.queries(this);
    container.mutations(this);

    final pEl = parent?.getContainerElement();

    if (pEl != null) {
      containers.inherit(pEl.containers);
      queryResolvers.inherit(pEl.queryResolvers);
      mutationResolvers.inherit(pEl.mutationResolvers);
      elements.inherit(pEl.elements);
    }

    containers[container.stateType] = container;
    queryResolvers.addAll(container.queryResolvers);
    mutationResolvers.addAll(container.mutationResolvers);
    elements[container.runtimeType] = this;

    container.stateNotifier.addListener(onChange);

    super.mount(parent, newSlot);
  }

  void subscribe(Element dependent, Object? aspect) {
    _subscribers.add(dependent);
    dependencyAspects[dependent] = aspect as QueryAspect;
    setDependencies(dependent, this);
  }

  @override
  void updateDependencies(Element dependent, Object? aspect) {
    subscribe(dependent, aspect);
  }

  void unsubscribe(Element element) {
    _subscribers.remove(element);
    dependencyAspects.remove(element);
  }

  @override
  void unmount() {
    _subscribers.clear();
    containers.clear();
    queryResolvers.clear();
    mutationResolvers.clear();
    elements.clear();

    container.stateNotifier.removeListener(onChange);
    _initSub?.cancel();
    container.shutDown();

    super.unmount();
  }

  onChange() {
    final u = _subscribers.fold(_Update(), maybeTriggerElementRebuild);

    u.dirty.forEach((element) {
      element.markNeedsBuild();
    });

    u.inactive.forEach((element) {
      _subscribers.remove(element);
      Effect.deactivate(element);
    });

    if (u.daemons.isNotEmpty) {
      WidgetsBinding.instance!.addPostFrameCallback((_) {
        u.daemons.forEach((e) {
          if (getDependencies(e.key) != this) {
            Effect.deactivate(e.key);
          } else {
            e.value.notify(e.key);
          }
        });
      });
    }
  }

  _Update maybeTriggerElementRebuild(_Update update, Element element) {
    if (getDependencies(element) != this) {
      return update..inactive.add(element);
    }

    final aspect = dependencyAspects[element];
    final updatePolicy = aspect!.resolveUpdatePolicy(container.stateNotifier);

    if (updatePolicy is Rebuild && updatePolicy.shouldRebuild) {
      return update..dirty.add(element);
    } else if (updatePolicy is NotifyListeners && updatePolicy.shouldNotify) {
      return update..daemons.add(MapEntry(element, updatePolicy));
    }

    return update;
  }
}
